@@ -54,27 +54,31 @@
 
 static int vt_switcher_pipe[2];
 
-static int drm_validate_connector_opt(
-    struct mp_log *log, const struct m_option *opt, struct bstr name,
-    struct bstr param);
+static int drm_connector_opt_help(
+    struct mp_log *log, const struct m_option *opt, struct bstr name);
+
+static int drm_mode_opt_help(
+    struct mp_log *log, const struct m_option *opt, struct bstr name);
 
 static int drm_validate_mode_opt(
     struct mp_log *log, const struct m_option *opt, struct bstr name,
-    struct bstr param);
+    const char **value);
 
 static void kms_show_available_modes(
     struct mp_log *log, const drmModeConnector *connector);
 
-static void kms_show_available_connectors(struct mp_log *log, int card_no);
+static void kms_show_available_connectors(struct mp_log *log, int card_no,
+                                          const char *card_path);
 static double mode_get_Hz(const drmModeModeInfo *mode);
 
 #define OPT_BASE_STRUCT struct drm_opts
 const struct m_sub_options drm_conf = {
     .opts = (const struct m_option[]) {
-        {"drm-connector", OPT_STRING_VALIDATE(drm_connector_spec,
-                                               drm_validate_connector_opt)},
-        {"drm-mode", OPT_STRING_VALIDATE(drm_mode_spec,
-                                          drm_validate_mode_opt)},
+        {"drm-device", OPT_STRING(drm_device_path), .flags = M_OPT_FILE},
+        {"drm-connector", OPT_STRING(drm_connector_spec),
+            .help = drm_connector_opt_help},
+        {"drm-mode", OPT_STRING_VALIDATE(drm_mode_spec, drm_validate_mode_opt),
+            .help = drm_mode_opt_help},
         {"drm-atomic", OPT_CHOICE(drm_atomic, {"no", 0}, {"auto", 1})},
         {"drm-draw-plane", OPT_CHOICE(drm_draw_plane,
             {"primary", DRM_OPTS_PRIMARY_PLANE},
@@ -122,6 +126,9 @@ static const char *connector_names[] = {
     "Virtual",   // DRM_MODE_CONNECTOR_VIRTUAL
     "DSI",       // DRM_MODE_CONNECTOR_DSI
     "DPI",       // DRM_MODE_CONNECTOR_DPI
+    "Writeback", // DRM_MODE_CONNECTOR_WRITEBACK
+    "SPI",       // DRM_MODE_CONNECTOR_SPI
+    "USB",       // DRM_MODE_CONNECTOR_USB
 };
 
 struct drm_mode_spec {
@@ -142,8 +149,15 @@ struct drm_mode_spec {
 static void get_connector_name(const drmModeConnector *connector,
                                char ret[MAX_CONNECTOR_NAME_LEN])
 {
-    snprintf(ret, MAX_CONNECTOR_NAME_LEN, "%s-%d",
-             connector_names[connector->connector_type],
+    const char *type_name;
+
+    if (connector->connector_type < MP_ARRAY_SIZE(connector_names)) {
+        type_name = connector_names[connector->connector_type];
+    } else {
+        type_name = "UNKNOWN";
+    }
+
+    snprintf(ret, MAX_CONNECTOR_NAME_LEN, "%s-%d", type_name,
              connector->connector_type_id);
 }
 
@@ -198,7 +212,8 @@ static bool setup_connector(struct kms *kms, const drmModeRes *res,
         connector = get_connector_by_name(kms, res, connector_name);
         if (!connector) {
             MP_ERR(kms, "No connector with name %s found\n", connector_name);
-            kms_show_available_connectors(kms->log, kms->card_no);
+            kms_show_available_connectors(kms->log, kms->card_no,
+                                          kms->primary_node_path);
             return false;
         }
     } else {
@@ -497,11 +512,62 @@ err:
     return false;
 }
 
-static int open_card(int card_no)
+static int open_card_path(const char *path)
 {
-    char card_path[128];
-    snprintf(card_path, sizeof(card_path), DRM_DEV_NAME, DRM_DIR_NAME, card_no);
-    return open(card_path, O_RDWR | O_CLOEXEC);
+    return open(path, O_RDWR | O_CLOEXEC);
+}
+
+static char *get_primary_device_path(struct mp_log *log, int *card_no)
+{
+    drmDevice *devices[DRM_MAX_MINOR] = { 0 };
+    int card_count = drmGetDevices2(0, devices, MP_ARRAY_SIZE(devices));
+    char *device_path = NULL;
+    bool card_no_given = (*card_no >= 0);
+
+    if (card_count < 0) {
+        mp_err(log, "Listing DRM devices with drmGetDevices failed! (%s)\n",
+               mp_strerror(errno));
+        goto err;
+    }
+
+    if (card_no_given && *card_no > (card_count - 1)) {
+        mp_err(log, "Card number %d given too high! %d devices located.\n",
+               *card_no, card_count);
+        goto err;
+    }
+
+    for (int i = card_no_given ? *card_no : 0; i < card_count; i++) {
+        drmDevice *dev = devices[i];
+
+        if (!(dev->available_nodes & (1 << DRM_NODE_PRIMARY))) {
+            if (card_no_given) {
+                mp_err(log,
+                       "DRM card number %d given, yet it does not have "
+                       "a primary node!\n", i);
+                break;
+            }
+
+            continue;
+        }
+
+        const char *primary_node_path = dev->nodes[DRM_NODE_PRIMARY];
+
+        mp_verbose(log, "Picked DRM card %d, primary node %s%s.\n",
+                   i, primary_node_path,
+                   card_no_given ? "" : " as the default");
+
+        device_path = talloc_strdup(log, primary_node_path);
+        *card_no = i;
+        break;
+    }
+
+    if (!device_path)
+        mp_err(log, "No primary DRM device could be picked!\n");
+
+err:
+    drmFreeDevices(devices, card_count);
+
+    return device_path;
 }
 
 static void parse_connector_spec(struct mp_log *log,
@@ -509,33 +575,54 @@ static void parse_connector_spec(struct mp_log *log,
                                  int *card_no, char **connector_name)
 {
     if (!connector_spec) {
-        *card_no = 0;
+        *card_no = -1;
         *connector_name = NULL;
         return;
     }
     char *dot_ptr = strchr(connector_spec, '.');
     if (dot_ptr) {
+        mp_warn(log, "Warning: Selecting a connector by index with drm-connector "
+                     "is deprecated. Use the drm-device option instead.\n");
         *card_no = atoi(connector_spec);
         *connector_name = talloc_strdup(log, dot_ptr + 1);
     } else {
-        *card_no = 0;
+        *card_no = -1;
         *connector_name = talloc_strdup(log, connector_spec);
     }
 }
 
-struct kms *kms_create(struct mp_log *log, const char *connector_spec,
+struct kms *kms_create(struct mp_log *log,
+                       const char *drm_device_path,
+                       const char *connector_spec,
                        const char* mode_spec,
                        int draw_plane, int drmprime_video_plane,
                        bool use_atomic)
 {
     int card_no = -1;
     char *connector_name = NULL;
+
     parse_connector_spec(log, connector_spec, &card_no, &connector_name);
+    if (drm_device_path && card_no != -1)
+        mp_warn(log, "Both DRM device and card number (as part of "
+                     "drm-connector) are set! Will prefer given device path "
+                     "'%s'!\n",
+                drm_device_path);
+
+    char *primary_node_path = drm_device_path ?
+                              talloc_strdup(log, drm_device_path) :
+                              get_primary_device_path(log, &card_no);
+
+    if (!primary_node_path) {
+        mp_err(log,
+               "Failed to find a usable DRM primary node!\n");
+        return NULL;
+    }
 
     struct kms *kms = talloc(NULL, struct kms);
     *kms = (struct kms) {
         .log = mp_log_new(kms, log, "kms"),
-        .fd = open_card(card_no),
+        .primary_node_path = primary_node_path,
+        .fd = open_card_path(primary_node_path),
         .connector = NULL,
         .encoder = NULL,
         .mode = {{0}},
@@ -551,12 +638,6 @@ struct kms *kms_create(struct mp_log *log, const char *connector_spec,
         goto err;
     }
 
-    char *devname = drmGetDeviceNameFromFd(kms->fd);
-    if (devname) {
-        mp_verbose(log, "Device name: %s\n", devname);
-        drmFree(devname);
-    }
-
     drmVersionPtr ver = drmGetVersion(kms->fd);
     if (ver) {
         mp_verbose(log, "Driver: %s %d.%d.%d (%s)\n", ver->name,
@@ -606,6 +687,7 @@ err:
         drmModeFreeResources(res);
     if (connector_name)
         talloc_free(connector_name);
+
     kms_destroy(kms);
     return NULL;
 }
@@ -652,12 +734,13 @@ static void kms_show_available_modes(
 }
 
 static void kms_show_foreach_connector(struct mp_log *log, int card_no,
+                                       const char *card_path,
                                        void (*show_fn)(struct mp_log*, int,
                                                        const drmModeConnector*))
 {
-    int fd = open_card(card_no);
+    int fd = open_card_path(card_path);
     if (fd < 0) {
-        mp_err(log, "Failed to open card %d\n", card_no);
+        mp_err(log, "Failed to open card %d (%s)\n", card_no, card_path);
         return;
     }
 
@@ -693,11 +776,13 @@ static void kms_show_connector_name_and_state_callback(
     mp_info(log, "  %s (%s)\n", other_connector_name, connection_str);
 }
 
-static void kms_show_available_connectors(struct mp_log *log, int card_no)
+static void kms_show_available_connectors(struct mp_log *log, int card_no,
+                                          const char *card_path)
 {
-    mp_info(log, "Available connectors for card %d:\n", card_no);
+    mp_info(log, "Available connectors for card %d (%s):\n", card_no,
+            card_path);
     kms_show_foreach_connector(
-        log, card_no, kms_show_connector_name_and_state_callback);
+        log, card_no, card_path, kms_show_connector_name_and_state_callback);
     mp_info(log, "\n");
 }
 
@@ -715,21 +800,45 @@ static void kms_show_connector_modes_callback(struct mp_log *log, int card_no,
     mp_info(log, "\n");
 }
 
-static void kms_show_available_connectors_and_modes(struct mp_log *log, int card_no)
+static void kms_show_available_connectors_and_modes(struct mp_log *log,
+                                                    int card_no,
+                                                    const char *card_path)
 {
-    kms_show_foreach_connector(log, card_no, kms_show_connector_modes_callback);
+    kms_show_foreach_connector(log, card_no, card_path,
+                               kms_show_connector_modes_callback);
 }
 
 static void kms_show_foreach_card(
-    struct mp_log *log, void (*show_fn)(struct mp_log*,int))
+    struct mp_log *log, void (*show_fn)(struct mp_log*,int,const char *))
 {
-    for (int card_no = 0; card_no < DRM_MAX_MINOR; card_no++) {
-        int fd = open_card(card_no);
-        if (fd < 0)
-            break;
-        close(fd);
-        show_fn(log, card_no);
+    drmDevice *devices[DRM_MAX_MINOR] = { 0 };
+    int card_count = drmGetDevices2(0, devices, MP_ARRAY_SIZE(devices));
+    if (card_count < 0) {
+        mp_err(log, "Listing DRM devices with drmGetDevices failed! (%s)\n",
+               mp_strerror(errno));
+        return;
     }
+
+    for (int i = 0; i < card_count; i++) {
+        drmDevice *dev = devices[i];
+
+        if (!(dev->available_nodes & (1 << DRM_NODE_PRIMARY)))
+            continue;
+
+        const char *primary_node_path = dev->nodes[DRM_NODE_PRIMARY];
+
+        int fd = open_card_path(primary_node_path);
+        if (fd < 0) {
+            mp_err(log, "Failed to open primary DRM node path %s!\n",
+                   primary_node_path);
+            continue;
+        }
+
+        close(fd);
+        show_fn(log, i, primary_node_path);
+    }
+
+    drmFreeDevices(devices, card_count);
 }
 
 static void kms_show_available_cards_and_connectors(struct mp_log *log)
@@ -747,31 +856,28 @@ double kms_get_display_fps(const struct kms *kms)
     return mode_get_Hz(&kms->mode.mode);
 }
 
-static int drm_validate_connector_opt(struct mp_log *log, const struct m_option *opt,
-                                      struct bstr name, struct bstr param)
+static int drm_connector_opt_help(struct mp_log *log, const struct m_option *opt,
+                                  struct bstr name)
 {
-    if (bstr_equals0(param, "help")) {
-        kms_show_available_cards_and_connectors(log);
-        return M_OPT_EXIT;
-    }
-    return 1;
+    kms_show_available_cards_and_connectors(log);
+    return M_OPT_EXIT;
+}
+
+static int drm_mode_opt_help(struct mp_log *log, const struct m_option *opt,
+                             struct bstr name)
+{
+    kms_show_available_cards_connectors_and_modes(log);
+    return M_OPT_EXIT;
 }
 
 static int drm_validate_mode_opt(struct mp_log *log, const struct m_option *opt,
-                                 struct bstr name, struct bstr param)
+                                 struct bstr name, const char **value)
 {
-    if (bstr_equals0(param, "help")) {
-        kms_show_available_cards_connectors_and_modes(log);
-        return M_OPT_EXIT;
-    }
-
-    char *spec = bstrto0(NULL, param);
-    if (!parse_mode_spec(spec, NULL)) {
+    const char *param = *value;
+    if (!parse_mode_spec(param, NULL)) {
         mp_fatal(log, "Invalid value for option drm-mode. Must be a positive number, a string of the format WxH[@R] or 'help'\n");
-        talloc_free(spec);
         return M_OPT_INVALID;
     }
-    talloc_free(spec);
 
     return 1;
 }
